#include <chrono>
#include <chronoptics/tof.hpp>
#include <iostream>
#include <string>
#include <vector>

namespace tof = chronoptics::tof;

struct Config {
  tof::CameraConfig camera_config;
  tof::ProcessingConfig processing_config;
  std::string name;
};

float measure_fps(tof::KeaCamera& cam, int seconds) {
  auto start = std::chrono::steady_clock::now();
  auto now = start;

  std::size_t count{0};
  while (now - start < std::chrono::seconds(seconds)) {
    cam.get_frames();
    count++;
    now = std::chrono::steady_clock::now();
  }

  return 1.0 * count / seconds;
}

std::vector<Config> create_configurations(tof::KeaCamera& cam,
                                          bool default_processing) {
  std::vector<Config> configurations;

  tof::ProcessingConfig processing_config{};
  processing_config.set_calibration_enabled(false);

  auto default_four_phase_steps = cam.get_camera_config();

  default_four_phase_steps.reset();
  default_four_phase_steps.set_phase_shifts(0, {0, 0.25, 0.5, 0.75});
  default_four_phase_steps.set_integration_time(0, {250, 250, 250, 250});
  default_four_phase_steps.set_mlx_mipi_speed(960);

  if (default_processing) {
    processing_config = default_four_phase_steps.default_processing();
  }

  configurations.push_back(Config{std::move(default_four_phase_steps),
                                  std::move(processing_config),
                                  "Four phase steps"});

  processing_config = tof::ProcessingConfig{};
  processing_config.set_calibration_enabled(false);

  auto four_phase_steps_qvga = cam.get_camera_config();

  four_phase_steps_qvga.reset();
  four_phase_steps_qvga.set_phase_shifts(0, {0, 0.25, 0.5, 0.75});
  four_phase_steps_qvga.set_integration_time(0, {250, 250, 250, 250});
  four_phase_steps_qvga.set_binning(0, 1);
  four_phase_steps_qvga.set_mlx_mipi_speed(960);

  if (default_processing) {
    processing_config = four_phase_steps_qvga.default_processing();
  }

  configurations.push_back(Config{std::move(four_phase_steps_qvga),
                                  std::move(processing_config),
                                  "Four phase steps qvga"});

  processing_config = tof::ProcessingConfig{};
  processing_config.set_calibration_enabled(false);
  processing_config.set_phase_unwrapping_enabled(true);

  auto dual_freq = cam.get_camera_config();
  dual_freq.reset();
  dual_freq.add_frame();

  dual_freq.set_phase_shifts(0, {0, 0.25, 0.5, 0.75});
  dual_freq.set_integration_time(0, {250, 250, 250, 250});

  dual_freq.set_phase_shifts(1, {0, 0.25, 0.5, 0.75});
  dual_freq.set_integration_time(1, {250, 250, 250, 250});

  dual_freq.set_mlx_mipi_speed(960);

  dual_freq.set_modulation_frequency(0, 80);
  dual_freq.set_modulation_frequency(1, 100);

  if (default_processing) {
    processing_config = dual_freq.default_processing();
  }

  configurations.push_back(Config{
      std::move(dual_freq), std::move(processing_config), "Dual frequency"});

  processing_config = tof::ProcessingConfig{};
  processing_config.set_calibration_enabled(false);
  processing_config.set_phase_unwrapping_enabled(true);

  auto dual_freq_qvga = cam.get_camera_config();
  dual_freq_qvga.reset();
  dual_freq_qvga.add_frame();

  dual_freq_qvga.set_phase_shifts(0, {0, 0.25, 0.5, 0.75});
  dual_freq_qvga.set_integration_time(0, {250, 250, 250, 250});

  dual_freq_qvga.set_phase_shifts(1, {0, 0.25, 0.5, 0.75});
  dual_freq_qvga.set_integration_time(1, {250, 250, 250, 250});

  dual_freq_qvga.set_modulation_frequency(0, 80);
  dual_freq_qvga.set_modulation_frequency(1, 100);

  dual_freq_qvga.set_binning(0, 1);
  dual_freq_qvga.set_binning(1, 1);
  dual_freq_qvga.set_mlx_mipi_speed(960);

  if (default_processing) {
    processing_config = dual_freq_qvga.default_processing();
  }

  configurations.push_back(Config{std::move(dual_freq_qvga),
                                  std::move(processing_config),
                                  "Dual frequency qvga"});

  return configurations;
}

int main(int argc, char** argv) {
  if (argc < 2 || argc > 5) {
    std::cout << "Expected " << argv[0]
              << " <serial> ?seconds? ?rgb? ?default_processing?\n";
    std::cout << "\tserial: The serial number of the camera.\n";
    std::cout << "\tseconds: The number of seconds to stream for, is by "
                 "default 10.\n";
    std::cout << "\trgb: Whether to stream rgb as well, by default disabled.\n";
    std::cout << "\tdefault_processing: Whether to do viewer default "
                 "processing, by default it does barebones.\n";
    return -1;
  }

  int seconds{10};
  if (argc >= 3) {
    seconds = atoi(argv[2]);
  }

  bool rgb{false};
  if (argc >= 4) {
    rgb = atoi(argv[3]);
  }

  bool default_processing{false};
  if (argc >= 5) {
    default_processing = atoi(argv[4]);
  }

  std::vector<tof::FrameType> types{tof::FrameType::Z};
  if (rgb) {
    types.push_back(tof::FrameType::BGR);
  }

  std::cout << "Doing frame rate test on camera " << argv[1] << " for "
            << seconds << " seconds, " << (rgb ? "" : "not ")
            << "streaming rgb, doing "
            << (default_processing ? "viewer default" : "barebones")
            << " processing\n";

  try {
    tof::ProcessingConfig processing_config;
    tof::KeaCamera cam(processing_config, argv[1]);

    auto original_config = cam.get_camera_config();

    auto configurations = create_configurations(cam, default_processing);

    for (auto& config : configurations) {
      std::cout << config.name << " fps:\n";
      cam.set_on_camera_processing(false);

      cam.set_camera_config(config.camera_config);
      cam.set_process_config(config.processing_config);
      tof::select_streams(cam, {types});

      cam.start();

      // Wait till we get first frame before measuring
      cam.get_frames();

      float fps = measure_fps(cam, seconds);
      std::cout << "\ton host:\t" << fps << "\n";

      cam.stop();

      cam.set_on_camera_processing(true);
      tof::select_streams(cam, {types});

      cam.start();

      cam.get_frames();

      fps = measure_fps(cam, seconds);
      std::cout << "\ton camera:\t" << fps << "\n";

      cam.stop();
    }

    tof::ProcessingConfig config;
    cam.set_process_config(config);
    cam.set_camera_config(original_config);

  } catch (std::exception& e) {
    std::cout << "Error: " << e.what() << "\n";
    return -1;
  }
}